/*
    slforum version 1.x - Simple chat room usable via telnet.
    Copyright (C) 1998-2001 Jeremy D. Monin, Dan Chin, Matt Lochansky
	(forum@shadowlands.org).
    Based on jdkchat version 1.3, J.D.Koftinoff Software, Ltd. (GPL'd)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#ifndef __SLFORUMUSER_H
#define	__SLFORUMUSER_H

typedef int userfield_t;  /* for !to-multiple */

/** NOTE: if you add new substates here, be sure to modify
 *  SLForumUser::admInputSubstateBehavTable to match.
 *  (in slforumuser-adminsubstate.inc.cc)
 */
typedef enum
{
  /* normally here */
  ss_nonadmin=0,

  /* admin menu */
  ss_mainmenu,

  /* add */
  ss_addgetuname, ss_addgetpw,    ss_addconfpw,
  ss_addgetgecos, ss_addaskadmin, ss_add_doit,

  /* edit */
  ss_editgetuname, ss_edit_retrieve,
  ss_editmenu,
  ss_edituname, ss_editconfuname, ss_edituname_doit,
  ss_editpw,    ss_editconfpw,    ss_editpw_doit,
  ss_editgecos, ss_editgecos_doit,
  ss_editisadmin, ss_edit_doit,

  /* del */
  ss_delgetuname, ss_dellookup, ss_delconfirm, ss_del_doit,

  ss_list,
  /* TODO TODO TODO - add LIST substates if any */

  /* virt users changing pw.. not really "admin" substates
   * but the substate mechanism set up for the admin modes
   * works nicely for it
   */
  ss_pwch_getoldpw, ss_pwch_getnewpw, ss_pwch_confnewpw, ss_pwch_setnewpw,

  /* macro: waiting for macro name key */
  ss_macro_expand_readkey,
  ss_macro_addedit_readkey,
  ss_macro_addedit_checkkey,
  ss_macro_addedit_askreplace,
  ss_macro_addedit_readline,
  ss_macro_addedit_ctrlq,     /* from _addedit_readline */
  ss_macro_addedit_checkline,
  ss_macro_addedit_askcrlf,
  ss_macro_addedit_doit,
  ss_macro_delete_readkey,
  ss_macro_delete_confirm,  /* see ss_macro_LAST below */

  SS_NUM_SUBSTATES, /* not an actual state, just counts 'em. */
  ss_unused	    /* marks an unused admInputSubstateBehavTable entry */

} admin_substate_t;

#define ss_macro_FIRST ss_macro_expand_readkey    /* first macro substate */
#define ss_macro_LAST  ss_macro_delete_confirm    /* last macro substate  */



/** This is for the table of admin substates; many of these substates
 *  simply need to get a line of text from the user, and move to a new
 *  substate when <return> is hit.  (setAdmSubstate does the processing
 *  for the state transition, including the bit of logic that actually
 *  needs to be done, such as new-password checking, etc.)  There are
 *  more complex admin substates; their behavior is handled by the
 *  switch within handleAdminLine.  The input-substate-behav-table
 *  handles the default: for that switch.
 */
typedef struct
{
  bool show_input;
  admin_substate_t nextstate;
} admin_input_substate_behav_t;



class SLForumRoom;
class SLForumUser;

#include "linkedlist.h"

typedef struct _slf_cmd_desc;

typedef union {
    char *strval;
    int   intval;
    SLForumUser *u;
    const struct _slf_cmd_desc *cmd;
} slfparam;


/*                            room           user       argc  argv       */
typedef int (slfusercmd) (SLForumRoom *, SLForumUser *, int, slfparam*);


typedef struct _slf_cmd_desc {
  const char *name;
  slfusercmd *meth;
  const char *parmspec;  // 19990529.1800
  const char *parmhelp;  // 19990212.0830
  const char *deschelp;	 // 19990212.0830
} slf_cmd_desc;



#include "slforumcmd.h"



/** If you add _any_ character-buffer fields to this class, put
 *  its size here as a constant, and use that constant,
 *  do _not_ do anything like
 *	mybuff[79] = '\0';
 *  or
 *	strncpy (mybuff, src, 79);
 *
 * (use USERMYBUFSIZ-1 instead)
 *
 *  Also be sure to add a line to the constructor to set
 *  the end of the string to \0; see constructor for why.
 */

#define USERBUFSIZ	80	// 1 line long
#define USERNAMESIZ	16
#define USERPROMPTSIZ	256	// should be dns size- 'prompt' is our
				// per-user scratch buffer as well.

#define USERAWAYSIZ	(80-12)	// used for away messages and comments,
				// which have the same format onscreen.

#define LONGFROMSIZ    256	// PAM wants whole thing, we might too later

#define DNSFROMSIZ	20	// since entire DNS-from string is too
				// long for the !who list format.



#ifdef ALLOW_MACROS

#define MACRO_RS  ((char) (30+128))
#define MACRO_SUB ((char) (26+128))

typedef struct {
    char key;
    bool  hit_enter_afterwards;
    const char* val;

    bool editing;  // true when editing this macro

    // within a val, ascii 30 (RS, record separator) separates lines.
    // ascii 26 (SUB) substitues where the command char goes.
    // in order to make life easier for anyone editing the file,
    // the high bit is set on these in-band control chars.

} macro_t;

#define macro_key(LL)  (((macro_t *)(LL))->key)
#define macro_val(LL)  (((macro_t *)(LL))->val)
#define macro_crlf(LL) (((macro_t *)(LL))->hit_enter_afterwards)
#define macro_editing(LL) (((macro_t *)(LL))->editing)

#endif  /* def(ALLOW_MACROS) */


class SLForumUser
{
public:
  //	SLForumUser( SLForumRoom *current_room);
	SLForumUser( SLForumRoom *current_room, int idx_in_room);
	~SLForumUser();

	friend class SLForumCmd;

	bool	isUserConnected() const;// valid network socket (not empty).
	bool	IsUserAccepted() const;	// validated, logged in, talking.
	bool	IsUserAround()   const;	// accepted() or waiting guest.
	bool	isAdmin()	 const; // admin.
	int	GetFD()          const;
	const char *GetName()    const;

	slf_tristate userCanLock()	const;

	bool	ManageUser();		// if rets false, connection has ended.
	int	Send( const char *text );	// send to user without FROM:
	int	Send( int msgdescflags, const char *from, const char *to,
		      const char *text);
		      		// send to user from user,
				// or public if to==NULL
	int	Send( int msgdescflags, const char *from, const char *to,
		      const char *text, int fromtominwid);
				// send to user from user, with a minimum
				// width for the from/to attribution field

		/* if hint not set, send text and set it.  This assumes
		 * we're sending a system message from SLFORUM.
		 * Optional msgdescflags.
		 */
	int	Send
	(const char *text, int hintbitset, bool wants_beep, int msgdescflags=0);

		/* send a system message from some user (SLFORUM, TIMER, etc)
	 	 * with msgdescflags and optional beep.
		 */
	int	Send
	(const char *fromsysusername,
	 int msgdescflags, const char *text, bool wants_beep);

		/* send a system message from SLFORUM, with msgdescflags;
		 * this allows the user to ignore certain types of message
		 * (such as users entering/leaving, or dice rolls) that come
		 * from the system, not from a user.  The msgdescflags should
		 * not have MSG_PUBLIC set.
		 */
	int	Send( int msgdescflags, const char *text, bool wants_beep );

		/* Check the whisperban list before sending. */
	bool	CheckBanned (SLForumUser *from);

	int	uWrite( const char *text);	// send bytes out to user_fd
	int	uWriteSubst( const char *text);	// send bytes w/ cmd char subst
	int	uWrite( const char *buf, int sz);  // send _some_ bytes out

	bool	isUserAway()  const;          // both accepted and away
	bool	isUserGuest() const;
	const char *setAway (const char *msg);
	const char *getAwayMessage () const;

	void	writeLogoutEntry (const char *fname);
	void	sendMessagesTo (SLForumUser *usr); // null for all
	void	sendMessagesAlso (SLForumUser *usr); // null for no effect :-)
	void	sendMessagesToNone ();
	void	sendMessagesFromNone ();

	int	idleTime() const;	// time since last keypress (secs)
	int	idleTime(time_t now) const;
	time_t	loginTime() const;	/* time of login (in time(2) form);
					 * use loginTimeStr() for string.
					 * The value returned by loginTime()
					 * is only meaningful for connected,
					 * authenticated users.
					 */
	void	pingAndCheckTimeout();
	bool	WasUserAccepted() const;	/* jer added 980814.0850 so we
					 * can see what they were doing when
					 * they left.  (This way, if someone
					 *  not-yet-successfully-logged-in
					 *  leaves, no "has left" message
					 *  is printed.)
					 * When user is in STATE_GET_LINE, return 1.
					 * Otherwise, return whether user was in
					 * STATE_GET_LINE when they left.
					 */

	const char *loginTimeStr() const;
	const char *loginFrom() const;	/* dns address logged in from; don't
					 * modify the value returned.
					 * This will be a string no longer
					 * than UT_HOSTSIZE in a utmp struct.
					 * Note that this may be larger than
					 * you'd think... Linux defines it
					 * as 256!
					 */
	const char *loginFrom16() const;
					/* Return the logged-in-from
					 * dns name, truncating if necessary
					 * to 16 chars (for !who list).
					 * jer added 980914.1015
					 */

	void talkingTo (int unum, bool stayte);
					/* for chatroom where the list of
					 * logged-in users is an indexed
					 * list.. STATE STATE STATE this
					 * is a temporary hack til we get
					 * ChatGroup objects together.
					 * (jer 19990113.1412)
					 */

	bool talkingTo(int unum) const;
	void talkingFrom (int unum, bool stayte);
	bool talkingFrom(int unum) const;

	/** Given the name of a command (without the initial '!' that marks it
	 *  as a command), return the parameters and one-line description.  If
	 *  the command isn't valid, return NULL for both of these.
	 * (jer 19990212.0830)
	 */
	void getHelpText (const char *cmdname,
			  const char **cmdparams, const char **cmddesc) const;
	void getHelpText (const char *cmdname, bool complete_partial,
			  const char **cmdparams, const char **cmddesc) const;

	int	getIgnoringMask() const; //19990606.0220 (removed
				   // isIgnoringBeeps of 19990318.0345)

	bool	WasUserManualDiscon() const; /* jer added 19990522.1515 so we
					 * can see if a user typed !exit, or if
					 * they were disconnected.
					 */
	void loadPrefs();
	int  savePrefs(); // 19991116.0200 dschin

	void lock_me();

	void setAutoMsgchk(int minutes);   // jer 20000508.2030

	void setAutoAway  (int minutes);   // jer 20000815.1530

#ifdef ALLOW_MACROS
	int	tryExpandMacro(char mc);  // jer 20001006.0010
	LinkedListLink* tryFindMacro
	    (char mc, LinkedListLink **LLprev=NULL) const;
	bool	tryAddMacro(macro_t *newmac);
	bool	tryDelMacro(char mc);
	int	macro_really_expand (const macro_t *mac);
	void	macro_append_line();
	void	listCurrentMacros();
	void	clearCurrentMacros();
	void	writeMacroToFile (FILE *f, macro_t *mac);
	macro_t *readMacroFromFile (FILE *f);
#endif  /* def(ALLOW_MACROS) */

	time_t lastping;	/* still there: jer added 19990702.1900 */

	int	idle_here_threshold;  /* if you're less than this many minutes
				       * idle and you're away, you're listed in
				       * the !here. (a user pref)
				       */

				/* for ignore fields; 19990606.0220 */
	static const int MSG_PUBLIC =   0x1;	// a public message
	static const int MSG_GUEST  =   0x2;	// from a guest
	static const int MSG_BEEP   =   0x4;	// beeping message
	static const int MSG_SYS    =   0x8;	// system message
	static const int MSG_MOVES  =  0x10;	// people arriving/leaving/away
	static const int MSG_BANNER =  0x20;	// it's a banner
	static const int MSG_ROLL   =  0x40;	// from c_roll command
	static const int MSG_P_BEEP =  0x80;	// private beep; c_beep command
	static const int MSG_HCHIME = 0x100;	// hourly chime
	static const int MSG_NOLOG  = 0x200;	// don't log this public line
	static const int MSG_EMOTE  = 0x400;    // emotes (!me)
	static const int MSG_BANNERED = 0x800;  // "banner from %s: %s"

	//Timer!  Timer consists of two arguments - number of minutes
        //and an optional message to go with it.
        //Added by MaTTeD 20000126.1815
	// note: if IsTimerSet == false but timer_expire_hour != -1,
	// there _was_ a timer set; timer_expire_hour and _min show
	// when it was set for.  This lets us see the previous timer
	// if we missed it going off.
	int		timer_expire_hour, timer_expire_min;
	char		timer_message[USERAWAYSIZ];
	bool		IsTimerSet;
	bool		IsTimerPublic;	// jer added 20000217.0343

protected:

	void 	ClearEntryLine();	// on screen, mind you
	void 	RedisplayEntryLine();	// also on screen
	void 	ClearBuffer();		// _this_ one clears the buffer content
	void	VisuallyClearBuffer();	// clear buf and show onscreen w/ ^H's


private:

	int	HandleNoConnection();
	int	HandleSendHello();
	int	HandleGetName();
	int	HandleGetPw();
	int	HandleFindDupl();
	int	HandleGetLine();
	int	HandleLockedLine();
	int	HandleAdminLine();
	int	HandleTelCmd();
	int	HandleTelOpt();

	int	slfgetch(char *ptrc);

	void	kick(const char *reason);
	void	rndCommand();	/* j added 980701.1640 */
	int	find_cmd_index(const char *buff, bool complete_partial) const;
		/* j added 19990126.2100 */

	void	tryPrintLoginBanner();

	char*	validateAndDowncaseName(char *buf);  /* j added 980805.2045 */
	/* _sendOut outputs a message from the user,
	 * to the whole room or privately (whispering).
	 * If to_bits == 0, _sendOut broadcasts,
	 * else it sends to the msgsTo user.
	 */
	void	_sendOut(char *sendthis);  /* j added 980701.1500 */

	void entry_extract_partial (char *partial, int &psize) const;
	void talkToMeToo (SLForumUser *newguy);
	void sendMssagesToNone();
	int idx_in_room() const;
	void    resetFrom();     // 19990224.2315 dschin

	void	try_unlock_me();
	slf_tristate user_can_lock;  // not everyone may have permission to.

	bool  show_input;

	struct	utmp uentry;  /* jer added 980524.2345 */
	time_t loginat;		/* login time: jer added 980702.0745 */
	time_t lastkey;		/* idle: jer added 980702.0745 */
	bool   all_keys_reset_idle;  /* if false, only public msgs reset idle;
				      * this is a flag used by the engine, not
				      * a user preference.
				      */
	bool	wants_true_idle; /* this is the user preference. (dflt: off) */
	bool	tab_expand_incl_yourname;  /* another user pref */
	bool	rot10_digits_too;	/* pref for c_rot13; defaults to yes */
	int	rot_offset;		/* pref for c_rot13; defaults to 13 */

public:   /* for SLFRoom to get at */
	int	auto_msgchk_minutes;	/* if nonzero, check every n minutes */
	time_t	auto_msgchk_next;	/* check next at this time */
	time_t	auto_msgchk_lastnew;	/* most recent new mail was here */
	bool	auto_msgchk_lastnew_is_valid;   /* if false, no new yet. */
	int	auto_away_seconds;	/* if nonzero, autoaway if this idle */

private:  /* back to normal */

	char ctstr[13];		/* login time string: jer added 980702.1800 */
	char longfrom[LONGFROMSIZ];     /* jer added 20000514.0730 */
	char dnsfrom[DNSFROMSIZ];	/* jer added 980914.1015, since the usual
				 * method of getting DNS is too big for !who
				 * list format.
				 */
	enum State
	{
		STATE_NO_CONNECTION,	/* When we go to this one, store the
					 * previous state in substate field,
					 * so we know what they were doing
					 * when they left.  -- jer 980814.0850
					 * (see WasUserAccepted() method)
					 */
		STATE_SETUP_CONN,	/* This state is to avoid races;
					 * it's set as soon as possible
					 * in HandleNoConnection().
					 * That method does connection init,
					 * then sets state to STATE_GET_NAME
					 * or STATE_FIND_DUPL.
					 */
		STATE_GET_NAME,
		STATE_JER_GETPW,
		STATE_FIND_DUPL,
		STATE_GET_LINE,		/* This is the only state which is
					 * considered 'accepted': the user is in
					 * the chatroom and typing (or may be !away).
					 */
		STATE_GUEST_CONFIRM,
		STATE_TEL_CMD, /* got IAC, waiting for WILL/WONT/etc (rfc854)*/
		STATE_TEL_OPT  /* got one o' those, waiting for opt char */
	} state;
	State prevstate;		/* mostly for logout; jer 981005.1015 */
	time_t substate;		/* for data within state; jer 980809.1020.  moved from 'int' to 'time_t' to support wider data (ctrl-d time) */
	admin_substate_t admin_substate; /* admin substate of STATE_GET_LINE */
	admin_substate_t admin_prev_substate;
	
	/** statedata use:
	 *
	 *  STATE_NO_CONNECTION: not used
	 *  STATE_GET_NAME:      not used
	 *  STATE_JER_GETPW:     not used
	 *
	 *  STATE_FIND_DUPL: holds SLForumUser* to more-idle self
	 *	which user is being prompted to kick or not
	 *
	 *  STATE_GET_LINE:
	 *	ss_nonadmin: not used
	 *	ss_mainmenu:?
	 *	ss_list:?
	 *	ss_addeditgetname:?
	 *	ss_addedit:?
	 *	ss_del:?
	 *	ss_pwch_getoldpw:?
	 *	ss_pwch_getnewpw:?
	 *	ss_pwch_confnewpw:?
	 *	ss_pwch_setnewpw: ?
	 *
	 *	ss_macro_addedit_readkey, _checkkey: a char cast to *
	 *	ss_macro_addedit_askreplace: * to the old macro struct
	 *	ss_macro_delete: * to old macro struct
	 *	    (and statedat2 is * to LLprev or NULL if first in list)
	 *
	 *  STATE_TEL_CMD, STATE_TEL_OPT:
	 *	these retain statedata from prev. states
	 *	for use when returning to that state after
	 *	handling the telnet command.
	 */
	void *statedata;	/* for data within state; jer 980809.1020 */
	void *statedat2;	/* so far only within ss_macro_addedit_* */
	time_t  waittil;	/* for bad-pw enforced wait; jer 981005.1025 */
	userfield_t to_bits;	/* for multiple !to; jer 19990113.1412 */
	userfield_t from_bits;	/* for multiple !to: jer 19990113.1412 */
	enum	n_talk_to_t {	/* for multiple !to: jer 19990622.1950 */
	  TALKING_TO_ALL,
	  TALKING_TO_ONE,
	  TALKING_TO_MANY
	} n_talking_to;

	int	talktomany_fieldwid;  /* 19990623.1830 */

	int     ignoringMask;	/* 19990606.0220 */
	int     blanklines;	/* for max_blank_lines: 19990606.0130 */
	char	cmd_char;	/* precede commands w/ this; normally '!' */
	bool	cmd_char_disabled;  /* ctrl-o toggles (for pasting stuff) */
	char	cmd_char_alt;	/* alternate command char, or \0  */
	bool	is_locked;	/* locked-out display mode */

	/** Read in one char, and place it in the input buffer (if it's
	 *  an input char) or act properly if it's a ctrl char.
	 */
	int	ManageInputBuffer();

	/** used by ManageInputBuffer: types this char into the buffer. */
	int	ManageInputNewChar(char c);

	/** used by ManageInputNewChar: try to handle this control char. */
	bool	ManageInputNewCtrlChar
		(char c, int &retval, bool admuserescape);

	/** used by HandleGetLine, and by
	 *  the handler for the admin substate,
	 *  since there's an escape from that substate
	 *  back to 'standard' STATE_GET_LINE.
	 */
	int	ManageChangedInputBuffer
	  (int manageinputbuffer_return_value);

	/* Handles displaying a full line or executing a command, and line
	 * wrap if needed.  (From ManageChangedInputBuffer)
	 */
	void	ManageFullLine (int ret);

	/* Handles user returning from being away.  (From
	 * ManageChangedInputBuffer)
	 */
	void	ManageUserReturn (int ret);

	int	HandleUserCommand(bool was_wrapped);
	void	resetFieldsToDiscon(bool manualDiscon=false);
	void	acceptAndWelcomeUser();
	void	setState(State newstate);	// jer added 981005.1015
	void	setState(State newstate, bool keepSubstate);	// jer added 981006.1350
	void	setAdmSubstate
		(admin_substate_t subst, bool justSetState=false);

	/** Set the prompt string; don't directly set the
	 *  contents of the prompt[] buffer, please... doing
	 *  it this way saves the need to strlen() every time
	 *  we word-wrap (which means every input char).
	 */
	void	setPrompt (const char *newprompt);
	/** Set the admin prompt to newprompt, and redisplay input line with
	 *  that prompt.  Call setAdmPrompt(NULL) to use the contents of
	 *  the user's non-admin slforumuser::prompt buffer.  The user's 
	 *  non-admin buffer will also be displayed if buf[0] == cmd_char.
	 *
	 *  Note that as a sideeffect, calling setAdmPrompt clears the buffer
	 *  before changing the prompt.
	 */
	void	setAdmPrompt (const char *newprompt);
	void	setPromptFromTo (bool always_do=false);
	int	getHint(int setbitfield) const;
	void	setHint(int setbitfield);

	/** Manipulate the listening-to list */ // 19990215.1900
	void	listenListReset();
	void	listenListAdd (const char *word);
	/** does the line contain anything in the list?
	 *  returns a pointer to the occurence, or NULL if none.
	 *  Don't use this unless you have to, it's not too efficient,
	 *  because ignore-public is not a common case.
	 *  endmatch, on return, points to the char after the matching word.
	 */
	char*	listenListMatch (const char *line, char **endmatch) const;

	/* set up listen list, name, load prefs, and check for admin...
	 * Just as things about entering the room go in AcceptAndWelcomeUser,
	 * things about logging in a real user go here.
	 */
	void	setupRealUserInfo (struct passwd *ppw);	// 20000201.1230


	// set up whisperban list	(20000610.1100)
	void	wbListReset();
	void	wbListAdd (const char *nam);
	bool	wbListMatch (const char *nam) const;


	// for ManageInputBuffer... takes a tab and makes it the right # of spaces.
	void expand_one_tab();
	// and goes back that # spaces ;-)
	void back_one_tab();

	// for max lines/sec brake.
	bool checkMaxLinesBrake();

	void pasteIntoBuffer
	    (const char *paste, int paste_strlen, int paste_displen,
	     bool paste_macro_input);

	void awaitTelCmd();


	/////////////////////////////////////////////////////////////
	// private fields
	//
	int 	cur_pos;  // position in array of end of line (next char here).
	int	cur_len;  // strlen of the buffer when displayed onscreen
	                  // (call stringDisplayLength to get it)
	bool	macro_input;  // if true, print special characters specially
	                      // (tab, ctrl-whatever, esc, etc.).  If false,
	                      // just echo buffer's contents to the user.
	char 	buffer[USERBUFSIZ];  // input buffer

	char	yankbuffer[USERBUFSIZ]; /* for ctrl-k/ctrl-y;
					 * jer added 19990811.1545 */
	int	yank_cur_pos, yank_cur_len;
	bool	yank_macro_input;  // if diff. then current macro_input while
			// pasting yankbuffer, need to recalc yank_cur_len.

	char	lastbuffer[USERBUFSIZ];  // jer 20000321.1400

	char 	name[USERNAMESIZ];
	char	prompt[USERPROMPTSIZ];  // to left of buffer
	int	promptlen;	/* jer added 980709.1715 */
	char	away[USERAWAYSIZ];  /* starts with the : or . after "is away";
				       jer added 980525.2045 */
	char	admprompt[USERBUFSIZ];	/* admin substate prompt */
	int	admpromptlen, realpromptlen;
	bool	local_mode;
	bool	is_away;	/* jer added 980525.2045 */
	time_t	away_at;	/* when went away -- dschin 19990922.1145 */
	bool	is_guest;	/* j added 980703.2330 */
	bool	is_admin;
	bool	is_virt;
	bool	running_cmd;	/* Are we running a command?  If so,
				 * nobody should mess with our input line,
				 * since we're doing that already.
				 * - j added 19990622.2050
				 */

  	struct 	sockaddr_in user_address;	
	int	user_fd;

	SLForumRoom *room;
	int my_idx_in_room;	/* where in the room array are we?
				 * jer 19990113.1412 */

	int hintbits;		/* bitfield for annoying help messages
				 * sent to user first time a feature is
				 * used, and not sent afterwards (like
				 * "type !to all" to end private message)
				 * -- not sent if corresp. field is 1.
				 * jer added 980709.1700 */

	char lastmfrom[USERNAMESIZ];
				/* who was the last message from?
				 * used for 19990215.0730 */
	char lastmto  [USERNAMESIZ];
				/* who was the last message to?
				 * used for 19990616.1500 */

	int lastm_rpt;		/* repeated number of lines from same user
				 * without an attribution @ left */

	LinkedList *listenlist;	/* words to listen for when we're
				 * ignoring public chat...
				 * 19990215.1900 */

	LinkedList *wblist;	/* people to ban whispering from...
				 * 19990610.1100 */


	/** for max lines per sec brake */

	int lines_per_sec[MAX_LINES_SECS_FOR_AVG];  /* 19990922.2100 */
	int lines_idx;	// curr. index into that array (last time we lined)
	time_t lastline;  // 'current' time; last time we wrote a line
	int lines_sum;  // sum of lines_per_sec array

	// for autowho, autotopic  -- dschin 19990922.1145
	int maxAwayIdle;	/* Max idle time before doing !who on return */
	int maxTopicIdle;	/* Max second before doing !topic on return  */

	// for macro_input
	char  macroNameBuf[6];

	char keystringChar(const char* keyster) const;
	const char* keyDisplayString (char mc, bool spaces_too=false);
	size_t keyDisplayLength (char mc, bool spaces_too=false);
	size_t stringDisplayLength
	    (const char *txt, int *strlenvar=NULL, bool spaces_too=false);

	// for macros
#ifdef ALLOW_MACROS
	LinkedList *macrolist;  /* elem data is of type macro_t */
	char macrostack[MAX_MACRO_CALL_LEVEL];
	int macro_next_level;   /* 1 to MAX_MACRO_CALL_LEVEL during
				 * macro execution, 0 when not running
				 * a macro.
				 */
#endif  /* def(ALLOW_MACROS) */


        static const char *welcome;	
	static const char *commands;
	static const char *rndCommands[];  /* jer added 980701.1640 */
	static const int NUM_RND_COMMANDS;
	static const int CONNECT_TIMEOUT;

 public:
	static const int HINT_TO_ALL = 1;  // type "!to all" to end priv msg
	static const int HINT_DID_TO_ALL = 2; // further msgs will sent to all
	static const int HINT_ACCEPT = 4;  // type "!accept" to...
	static const int HINT_KICK = 8;  // type "!kick" to...
	static const int HINT_AWAY = 16;  // you're marked as away...
	static const int HINT_PREVTOPIC = 32;  // type !prevtopic n to get...
	static const int HINT_SHOWAWAY  = 64;  // type !showaway to list away..
	static const int HINT_HCHIME  = 128;   // type !ignore chime ...
	static const int HINT_CCDISAB = 256;   // cmd char disabled, hit...
	static const int HINT_IGN_BANNERED = 512;  // type !ignore bannered ...

	static const int c_toall_idx;
	static const slf_cmd_desc usercmds[];
	static const admin_input_substate_behav_t admInputSSBehavTable[];

	// for debug -- jer added 20001001
	void debug_printSlotStatus () const;
};

#endif	// __SLFORUMUSER_H
